#from .Flow import *
from .LAMINAR import LAMINAR
from .utils import standardize


def add_one(x: int):
    """An example function that increases a number

    :param x: The input parameter to increase
    :return: The successor of the given number
    """
    return x + 1
